from dataclasses import dataclass

# from acslib.ccure.base import CcureACS
# from acslib.base.connection import ACSRequestData


@dataclass
class V2Endpoints:
    FIND_OBJS_W_CRITERIA = "/victorwebservice/api/Objects/FindObjsWithCriteriaFilter"
    CLEARANCES_FOR_ASSIGNMENT = "/victorwebservice/api/v2/Personnel/ClearancesForAssignment"
    GET_ALL_WITH_CRITERIA = "/victorwebservice/api/Objects/GetAllWithCriteria"
    PERSIST_TO_CONTAINER = "/victorwebservice/api/Objects/PersistToContainer"
    REMOVE_FROM_CONTAINER = "/victorwebservice/api/Objects/RemoveFromContainer"
    LOGIN = "/victorwebservice/api/Authenticate/Login"
    LOGOUT = "/victorwebservice/api/Authenticate/Logout"
    KEEPALIVE = "/victorwebservice/api/v2/session/keepalive"
    VERSIONS = "/victorwebservice/api/Generic/Versions"
    DISABLE = "/victorwebservice/api/v2/objects/SetProperty"


# class FindObjsWithCriteria(CcureACS):
#     def __init__(self, connection, **kwargs):
#         super().__init__(connection)
#         self.type_full_name = kwargs.get("type_full_name")
#         self.page_size = kwargs.get("page_size", self.connection.config.page_size)
#         self.page_number = kwargs.get("page_number", 1)
#         self.where_clause = kwargs.get("where_clause", "")



from enum import Enum
from unittest.mock import patch

import pytest

from acslib.ccure.base import CcureACS, CcureConnection
from acslib.ccure.search import PersonnelFilter, SearchTypes


def test_default_ccure_acs(env_config, caplog):
    """Default picks up env vars."""
    ccure = CcureACS()
    assert ccure.config.base_url == "https://example.com/ccure"
    assert ccure.logger.name == "acslib.ccure.base"
    assert "acslib.ccure.base" in caplog.text


def test_user_supplied_logger(env_config, caplog):
    """."""
    import logging

    cc_conn = CcureConnection(logger=logging.getLogger("test"))
    ccure = CcureACS(connection=cc_conn)
    assert ccure.logger.name == "test"
    assert "test:base" in caplog.text


def test_default_ccure_search(env_config, personnel_response, caplog):
    ccure = CcureACS()
    with patch(
        "acslib.ccure.base.CcureACS._search_people", return_value=personnel_response
    ) as mock_search:
        ccure.search(search_type=SearchTypes.PERSONNEL, terms=["test"])
        mock_search.assert_called_with(["test"])
    assert "Searching for personnel" in caplog.text


def test_ccure_search_with_filter(env_config, personnel_response, caplog):
    ccure = CcureACS()
    filter = PersonnelFilter()
    with patch(
        "acslib.ccure.base.CcureACS._search_people", return_value=personnel_response
    ) as mock_search:
        ccure.search(search_type=SearchTypes.PERSONNEL, terms=["test"], search_filter=filter)
        mock_search.assert_called_with(["test"], filter)
    assert "Searching for personnel" in caplog.text


def test_invalid_search_type(env_config):
    class NewTypes(Enum):
        NEW = "new"
        PERSONNEL = "personnel"

    ccure = CcureACS()
    with pytest.raises(ValueError):
        ccure.search(search_type=NewTypes.NEW, terms=["test"])



from enum import Enum

from acslib.base.search import ACSFilter, BooleanOperators, TermOperators


def left_fuzz(term):
    return f"%{term}"


def right_fuzz(term):
    return f"{term}%"


def full_fuzz(term):
    return f"%{term}%"


def no_fuzz(term):
    return f"{term}"


LFUZZ = left_fuzz
RFUZZ = right_fuzz
FUZZ = full_fuzz
NFUZZ = no_fuzz

PERSONNEL_LOOKUP_FIELDS = {"FirstName": FUZZ, "LastName": FUZZ}
CLEARANCE_LOOKUP_FIELDS = {"Name": FUZZ}


class SearchTypes(Enum):
    PERSONNEL = "personnel"
    CLEARANCE = "clearance"


class BaseCcureFilter(ACSFilter):
    """Base CCure Filter
    :param lookups: List of tuples containing the field name and the lookup function
    :param outer_bool: Boolean operator to use between search terms
    :param inner_bool: Boolean operator to use between lookups
    :param term_operator: Term operator to use between field and a search term
    :attribute
    """

    def __init__(
        self,
        lookups: list[tuple[str, callable]] = None,
        outer_bool=BooleanOperators.AND,
        inner_bool=BooleanOperators.OR,
        term_operator=TermOperators.FUZZY,
    ):
        self.filter_fields = lookups
        self.outer_bool = outer_bool.value
        self.inner_bool = inner_bool.value
        self.term_operator = term_operator.value
        #: List of properties from CCURE to be included in the CCURE response
        self.display_properties = ["FirstName", "MiddleName", "LastName", "ObjectID"]

    def _compile_term(self, term: str) -> str:
        """Get all parts of the query for one search term"""
        fields = [(field_name, lookup(term)) for field_name, lookup in self.filter_fields.items()]
        field_queries = [f"{field_name} {self.term_operator} '{lookup}'" for field_name, lookup in fields]
        return f"({self.inner_bool.join(field_queries)})"

    def update_display_properties(self, properties: list[str]):
        if not isinstance(properties, list):
            raise TypeError("Properties must be a list of strings")
        self.display_properties += properties

    def filter(self, search):
        pass


class PersonnelFilter(BaseCcureFilter):
    """Basic CCure Personnel Filter
    :param lookups: List of tuples containing the field name and the lookup function
    :param outer_bool: Boolean operator to use between search terms
    :param inner_bool: Boolean operator to use between lookups
    :param term_operator: Term operator to use between field and a search term
    :attribute
    """

    def __init__(
        self,
        lookups: dict[str, callable] = [],
        outer_bool=BooleanOperators.AND,
        inner_bool=BooleanOperators.OR,
        term_operator=TermOperators.FUZZY,
    ):
        self.filter_fields = lookups if lookups else PERSONNEL_LOOKUP_FIELDS
        self.outer_bool = f" {outer_bool.value} "
        self.inner_bool = f" {inner_bool.value} "
        self.term_operator = term_operator.value
        self.display_properties = ["FirstName", "MiddleName", "LastName"]

    def filter(self, search: list[str]) -> str:
        if not isinstance(search, list):
            raise TypeError("Search must be a list of strings")
        return self.outer_bool.join(self._compile_term(term) for term in search)


class ClearanceFilter(BaseCcureFilter):
    """Basic CCure Clearance Filter
    :param lookups: List of tuples containing the field name and the lookup function
    :param outer_bool: Boolean operator to use between search terms
    :param inner_bool: Boolean operator to use between lookups
    :param term_operator: Term operator to use between field and a search term
    :attribute
    """

    def __init__(
        self,
        lookups: dict[str, callable] = [],
        outer_bool=BooleanOperators.AND,
        inner_bool=BooleanOperators.OR,
        term_operator=TermOperators.FUZZY,
    ):
        self.filter_fields = lookups if lookups else CLEARANCE_LOOKUP_FIELDS
        self.outer_bool = f" {outer_bool.value} "
        self.inner_bool = f" {inner_bool.value} "
        self.term_operator = term_operator.value
        # List of properties from CCURE to be included in the CCURE response
        self.display_properties = ["Name"]

    def filter(self, search: list[str]) -> str:
        if not isinstance(search, list):
            raise TypeError("Search must be a list of strings")
        return self.outer_bool.join(self._compile_term(term) for term in search)



import logging
from numbers import Number
from typing import Optional

from acslib.base import (
    AccessControlSystem,
    ACSConnection,
    ACSRequestData,
    ACSRequestException,
    ACSRequestResponse,
    status,
)
from acslib.base.connection import ACSRequestMethod
from acslib.base.search import ACSFilter
from acslib.ccure.config import CcureConfigFactory
from acslib.ccure.search import PersonnelFilter, ClearanceFilter, SearchTypes, FUZZ

logger = logging.getLogger(__name__)


class CcureConnection(ACSConnection):
    def __init__(self, **kwargs):
        """
        A connection object to the CCure Server.
        Parameters:
        :param kwargs:
        """
        self.session_id = None
        if con_logger := kwargs.get("logger"):
            self.logger = con_logger
        else:
            self.logger = logger
        if not kwargs.get("config"):
            kwargs["config"] = CcureConfigFactory()
        self.logger.info("Initializing CCure connection")
        super().__init__(**kwargs)

    @property
    def headers(self):
        """."""
        return {
            "session-id": self.get_session_id(),
            "Access-Control-Expose-Headers": "session-id",
        }

    def login(self):
        """."""
        try:
            response = self.request(
                ACSRequestMethod.POST,
                request_data=ACSRequestData(
                    url=self.config.base_url + self.config.endpoints.LOGIN,
                    data=self.config.connection_data,
                ),
            )
            self.session_id = response.headers["session-id"]
            self.logger.debug(f"Fetched new Session ID: {self.session_id}")
        except ACSRequestException as e:
            self.logger.error(f"Error Fetching Session ID: {e}")
            self.log_session_details()
            self.logger.debug(f"Connection data: {self.config.connection_data}")
            raise e
        return self.session_id

    def logout(self):
        """Log out of the CCure session"""
        if self.session_id:
            self.logger.debug(f"Logging out of CCure session: {self.session_id}")
            try:
                self.request(
                    ACSRequestMethod.POST,
                    request_data=ACSRequestData(
                        url=self.config.base_url + self.config.endpoints.LOGOUT,
                        headers={"session-id": self.session_id},
                    ),
                )
            except ACSRequestException as e:
                self.logger.error(f"Error logging out of CCure session: {e}")
                self.log_session_details()
            finally:
                self.logger.debug(f"Removing Session ID: {self.session_id}")
                self.session_id = None

    def get_session_id(self):
        """."""
        self.logger.debug(f"Session ID: {self.session_id}")
        if self.session_id is None:
            return self.login()
        return self.session_id

    def keepalive(self):
        """
        Prevent the CCure api session from expiring from inactivity.
        """
        self.logger.debug(f"Keeeping CCure session alive: {self.session_id}")
        try:
            self.request(
                ACSRequestMethod.POST,
                request_data=ACSRequestData(
                    url=self.config.base_url + self.config.endpoints.KEEPALIVE,
                    headers={
                        "session-id": self.get_session_id(),
                        "Access-Control-Expose-Headers": "session-id",
                    },
                ),
            )
            self.logger.debug(f"Session kept alive: {self.session_id}")
        except ACSRequestException as e:
            self.logger.error(f"Error keeping CCure session alive: {e}")
            self.log_session_details()
            self.logout()

    def request(
        self,
        requests_method: ACSRequestMethod,
        request_data: ACSRequestData,
        timeout: Number = 0,
        request_attempts: int = 2,
    ) -> ACSRequestResponse:
        """
        Call the `ACSConnection.handle_requests` function and return the result.
        If the response is a 401, get a new CCure session_id and try the request again.

        Parameters:
            requests_method: A method from the requests module. get, post, etc
            request_data: Data used as kwargs for the requests_method
            timeout: Maximum time to wait for a server response, in seconds
            request_attempts: Maximum number of times to try the request

        Returns: An object with status_code, json, and headers attributes
        """
        # use TIMEOUT as the default timeout value
        if timeout != 0:
            self.config.timeout = timeout
        while request_attempts > 0:
            try:
                return super().request(requests_method, request_data)
            except ACSRequestException as e:
                if e.status_code != status.HTTP_401_UNAUTHORIZED or request_attempts == 1:
                    raise e
                request_attempts -= 1
                self.logout()
                request_data.headers["session-id"] = self.get_session_id()

    def log_session_details(self):
        """Log session ID and the api version number"""
        version_url = self.config.base_url + self.config.endpoints.VERSIONS
        self.logger.error(f"Session ID: {self.session_id}")
        try:
            response = self.request(
                ACSRequestMethod.POST,
                request_data=ACSRequestData(url=version_url),
            ).json
            self.logger.debug(f"CCure webservice version: {response.get('webServiceVersion')}")
            self.logger.debug(f"CCure app server version: {response.get('appServerVersion')}")
        except ACSRequestException as e:
            self.logger.debug(f"Could not get CCure api version number: {e}")


class CcureACS(AccessControlSystem):
    """."""

    def __init__(self, connection: Optional[CcureConnection] = None):
        """."""
        super().__init__(connection=connection)
        if not self.connection:
            self.connection = CcureConnection()
        self.logger = self.connection.logger
        self.request_options = {}
        self.search_filter = None

    @property
    def config(self):
        """."""
        return self.connection.config

    def _search_people(self, terms: list, search_filter: Optional[ACSFilter] = None) -> ACSRequestResponse:
        if not search_filter:
            search_filter = PersonnelFilter()
        request_json = {
            "TypeFullName": "Personnel",
            "DisplayProperties": search_filter.display_properties,
            "pageSize": self.connection.config.page_size,
            "pageNumber": 1,
            "WhereClause": search_filter.filter(terms),
        }
        if not search_filter.display_properties:
            del request_json["DisplayProperties"]

        return self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.connection.config.base_url
                + self.connection.config.endpoints.FIND_OBJS_W_CRITERIA,
                request_json=request_json,
                headers=self.connection.headers,
            ),
        )

    def _search_clearances(self, terms: list, search_filter: Optional[ACSFilter] = None) -> ACSRequestResponse:
        if not search_filter:
            search_filter = ClearanceFilter()
        request_json = {
            "partitionList": [],
            "propertyList": search_filter.display_properties,
            "pageSize": self.connection.config.page_size,
            "pageNumber": 1,
            "whereClause": search_filter.filter(terms),
            "sortColumnName": "",
            "whereArgList": [],
            "explicitPropertyList": [],
        }
        if not search_filter.display_properties:
            del request_json["DisplayProperties"]

        response = self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.connection.config.base_url
                + self.connection.config.endpoints.CLEARANCES_FOR_ASSIGNMENT,
                request_json=request_json,
                headers=self.connection.headers,
            ),
        )
        if response.json:
            response.json = response.json[1:]

        return response

    def search(
        self, search_type: SearchTypes, terms: list, search_filter: Optional[ACSFilter] = None
    ) -> ACSRequestResponse:
        match search_type:
            case SearchTypes.PERSONNEL.value:
                self.logger.info("Searching for personnel")
                return self._search_people(terms, search_filter)
            case SearchTypes.CLEARANCE.value:
                self.logger.info("Searching for clearances")
                return self._search_clearances(terms, search_filter)
            case _:
                raise ValueError(f"Invalid search type: {search_type}")



from acslib.base import ACSRequestResponse
from acslib.base.connection import ACSRequestData, ACSRequestMethod
from acslib.ccure.base import CcureACS
from acslib.ccure.search import ClearanceFilter, PersonnelFilter


class CCurePersonnel(CcureACS):
    def __init__(self, connection):
        super().__init__(connection)
        self.request_options = {
            "TypeFullName": "Personnel",
            "pageSize": self.connection.config.page_size,
            "pageNumber": 1,
        }
        self.search_filter = PersonnelFilter()

    def search(self, terms: list, search_filter: PersonnelFilter = None) -> ACSRequestResponse:
        self.logger.info("Searching for personnel")
        if search_filter:
            self.search_filter = search_filter
        request_json = {
            "DisplayProperties": self.search_filter.display_properties,
            "WhereClause": self.search_filter.filter(terms),
        }
        request_json.update(self.request_options)
        if not self.search_filter.display_properties:
            del request_json["DisplayProperties"]

        return self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.config.base_url + self.config.endpoints.FIND_OBJS_W_CRITERIA,
                request_json=request_json,
                headers=self.connection.headers,
            ),
        )

    def count(self) -> int:
        self.request_options["pageSize"] = 0
        self.request_options["CountOnly"] = True
        self.request_options["WhereClause"] = ""
        return self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.config.base_url + self.config.endpoints.FIND_OBJS_W_CRITERIA,
                request_json=self.request_options,
                headers=self.connection.headers,
            ),
        ).json

    def update(self, record_id: str, update_data: dict) -> ACSRequestResponse:
        pass

    def create(self, create_data: dict) -> ACSRequestResponse:
        pass

    def delete(self, record_id: str) -> ACSRequestResponse:
        pass


class CCureClearance(CcureACS):
    def __init__(self, connection):
        super().__init__(connection)
        self.request_options = {
            "partitionList": [],
            "pageSize": self.connection.config.page_size,
            "pageNumber": 1,
            "sortColumnName": "",
            "whereArgList": [],
            "propertyList": ["Name"],
            "explicitPropertyList": [],
        }
        self.search_filter = ClearanceFilter()

    def search(self, terms: str, search_filter: ClearanceFilter = None) -> ACSRequestResponse:
        self.logger.info("Searching for clearances")
        if search_filter:
            self.search_filter = search_filter
        request_json = {
            "whereClause": self.search_filter.filter(terms),
        }
        request_json.update(self.request_options)
        return self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.connection.config.base_url
                + self.connection.config.endpoints.CLEARANCES_FOR_ASSIGNMENT,
                request_json=request_json,
                headers=self.connection.headers,
            ),
        )

    def count(self) -> int:
        request_options = {}
        request_options["TypeFullName"] = "Clearance"
        request_options["pageSize"] = 0
        request_options["pageNumber"] = 1
        request_options["CountOnly"] = True
        request_options["WhereClause"] = ""

        return self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.config.base_url + self.config.endpoints.FIND_OBJS_W_CRITERIA,
                request_json=request_options,
                headers=self.connection.headers,
            ),
        ).json

    def update(self, record_id: str, update_data: dict) -> ACSRequestResponse:
        pass

    def create(self, create_data: dict) -> ACSRequestResponse:
        pass

    def delete(self, record_id: str) -> ACSRequestResponse:
        pass
